/*
	(c) 2021 Dmitry Grinberg   https://dmitry.gr
	Non-commercial use only OR licensing@dmitry.gr
*/

#include <string.h>
#include "usbPublic.h"
#include "timebase.h"
#include "usbDev.h"
#include "printf.h"
#include "usart.h"
#include "dz11.h"

#pragma GCC optimize ("Os")


#define MY_VID							('D' + ('G' << 8))	//<voice type="hippie">You can't own numbers, man...</voice>
#define MY_PID							('L' + ('C' << 8))

#define STR_IDX_LANG_ID					0
#define STR_IDX_MANUF					1
#define STR_IDX_PRODNAME				2
#define STR_IDX_SNUM					3

#define CFG_IDX_UARTS					1

#define NUM_CDCS						2		//windows gets pissy if more

#define IF_NO_OFST_CDC_CTRL				0
#define IF_NO_OFST_CDC_DATA				1

#define EP_NO_OFST_CDC_CTRL				1		//1 IN
#define EP_SZ_CDC_CTRL					8
#define EP_NO_OFST_CDC_DATA_IN			2		//2 IN
#define EP_SZ_CDC_DATA_IN				8
#define EP_NO_OFST_CDC_DATA_OUT			2		//2 OUT
#define EP_SZ_CDC_DATA_OUT				8


//RX state
static volatile uint8_t __attribute__((aligned(4))) mRxBufLL[NUM_CDCS][EP_SZ_CDC_DATA_OUT];
static volatile uint8_t mRxQueued[NUM_CDCS];

//TX state
static volatile bool mRunning;




static void usartPrvHandleRx(uint_fast8_t cdcIdx)
{
	uint_fast8_t lineNo = 3 - cdcIdx, epNo = EP_NO_OFST_CDC_DATA_OUT + 2 * cdcIdx;
	uint_fast8_t i, j, L = mRxQueued[cdcIdx];

	for (i = 0; i < L && dz11numBytesFreeInRxBuffer(lineNo); i++)
		dz11charRx(lineNo, mRxBufLL[cdcIdx][i]);
	
	if (i < L) {	//more chars left
		
		for (j = 0; i < L; i++, j++)
			mRxBufLL[cdcIdx][j] = mRxBufLL[cdcIdx][i];
		mRxQueued[cdcIdx] = j;
	}
	else if (L) {	//we HAD chars but no longer do
		
		mRxQueued[cdcIdx] = 0;
		usbRxRelease(epNo);
	}
}

extern void dz11rxSpaceNowAvail(uint_fast8_t line)
{
	uint_fast8_t cdcIdx = 3 - line;
	
	usartPrvHandleRx(cdcIdx);
}

void usbExtEpDataArrivalNotif(uint8_t epNo, uint_fast8_t len)
{
	uint_fast8_t cdcIdx = (epNo - EP_NO_OFST_CDC_DATA_OUT) / 2;
	
	if (epNo < EP_NO_OFST_CDC_DATA_OUT || ((epNo - EP_NO_OFST_CDC_DATA_OUT) % 2) || cdcIdx >= NUM_CDCS) {
		
		usbRxRelease(epNo);
		return;
	}
	
	mRxQueued[cdcIdx] = len;
	usartPrvHandleRx(cdcIdx);
}	

void usbExtStateChangedNotif(enum UsbState nowState)
{
	mRunning = nowState == UsbRunning;
}

void usbExtConfigSelected(uint8_t cfg)
{
	if (cfg == CFG_IDX_UARTS) {
		
		uint_fast8_t i;
		
		for (i = 0; i < NUM_CDCS; i++) {
		
			if (!usbEpIntrCfgIn(EP_NO_OFST_CDC_CTRL + 2 * i, EP_SZ_CDC_CTRL))
				pr("failed to configure (1 + 2 * %u).IN\n", i);
			if (!usbEpBulkCfgIn(EP_NO_OFST_CDC_DATA_IN + 2 * i, EP_SZ_CDC_DATA_IN))
				pr("failed to configure (2 + 2 * %u).IN\n", i);
			if (!usbEpBulkCfgOut(EP_NO_OFST_CDC_DATA_OUT + 2 * i, EP_SZ_CDC_DATA_OUT, (void*)mRxBufLL[i]))
				pr("failed to configure (2 + 2 * %u).OUT\n", i);
		}
	}
	else {
		
		//deconfigured... not really supported
	}
}

bool usbExtEp0handler(const struct UsbSetup *setup, void **dataOutP, int16_t *lenOutP)
{
	//as long as we only have CDCs, this is easy :D
	
	if ((setup->bmRequestType & 0x7f) == 0x21) {
		
		const uint32_t myBaudRate = 9600;
		
		if (!(setup->bmRequestType & 0x80) && setup->bRequest == 0x20 && setup->wLength == 7) {		//SET_LINE_CODING	(sz 7)
			
			uint_fast8_t len;
			uint8_t data[7];
			
			len = usbEp0dataRx(data);
			
			(void)len;
			//we got 7 bytes of line coding
			
			*lenOutP = 0;		//suuure...
			return true;
		}
		else if (!(setup->bmRequestType & 0x80) && setup->bRequest == 0x22) {	//SET_LINE_CONTROL_STATE
			
			*lenOutP = 0;		//suuure...
			return true;
		}
		else if ((setup->bmRequestType & 0x80) && setup->bRequest == 0x21 && setup->wLength == 7) {	//GET_LINE_CODING	(sz 7)
			
			uint8_t *dataP = *dataOutP;
			
			*dataP++ = (myBaudRate >> 0) & 0xff;
			*dataP++ = (myBaudRate >> 8) & 0xff;
			*dataP++ = (myBaudRate >> 16) & 0xff;
			*dataP++ = (myBaudRate >> 24) & 0xff;
			*dataP++ = 0;	//1 stop bit
			*dataP++ = 8;	//no parity
			*dataP++ = 8;	//8 data bits
			*lenOutP = 7;
			
			return true;
		}
		
		*lenOutP = -1;
		return true;
	}
	return false;
}

//descriptors
static const struct UsbDeviceDescriptor mDevDescr = {
	.bLength = sizeof(struct UsbDeviceDescriptor),
	.bDescriptorType = DESCR_TYP_DEVICE,
	.bcdUSB = 0x0200,
	.bDeviceClass = 0xef,
	.bDeviceSubClass = 0x02,
	.bDeviceProtocol = 0x01,
	.bMaxPacketSize0 = EP0_SZ,
	.idVendor = MY_VID,
	.idProduct = MY_PID,
	.bcdDevice = 0x0100,
	.iManufacturer = STR_IDX_MANUF,
	.iProduct = STR_IDX_PRODNAME,
	.iSerialNumber = STR_IDX_SNUM,
	.bNumConfigurations = 1,
};

static const struct {
	struct UsbConfigDescriptor cfg;
	struct {
		struct IfaceAssocDescriptor assocDescr;
		struct UsbInterfaceDescriptor ifCdcCtrl;
		struct UsbCdcHeaderFunctionalDescriptor ifCdcHdr;
		struct UsbCdcAcmFunctionalDescriptor ifCdcAcm;
		struct UsbCdcUnionFunctionalDescriptor ifCdcUnion;
		struct UsbCdcCallManagementFunctionalDescriptor ifCdcCallMgmnt;
		struct UsbEndpointDescriptor epCdcCtrl;
		struct UsbInterfaceDescriptor ifCdcData;
		struct UsbEndpointDescriptor epCdcDataOut;
		struct UsbEndpointDescriptor epCdcDataIn;
	} __attribute__((packed)) cdc[NUM_CDCS];
} __attribute__((packed)) mCfgDescr = {
	.cfg = {
		.bLength = sizeof(struct UsbConfigDescriptor),
		.bDescriptorType = DESCR_TYP_CONFIG,
		.numBytesReturned =	sizeof(struct UsbConfigDescriptor) +
								NUM_CDCS * (
									sizeof(struct IfaceAssocDescriptor) +
									sizeof(struct UsbInterfaceDescriptor) +
									sizeof(struct UsbCdcHeaderFunctionalDescriptor) +
									sizeof(struct UsbCdcAcmFunctionalDescriptor) +
									sizeof(struct UsbCdcUnionFunctionalDescriptor) +
									sizeof(struct UsbCdcCallManagementFunctionalDescriptor) +
									sizeof(struct UsbEndpointDescriptor) +
									sizeof(struct UsbInterfaceDescriptor) +
									sizeof(struct UsbEndpointDescriptor) +
									sizeof(struct UsbEndpointDescriptor)
								),
		.numIfaces = 2 * NUM_CDCS,	//2 per CDC instance
		.thisCfgIdx = CFG_IDX_UARTS,
		.bAttributes = 0x80,
		.bCurrent = 50,
	},
	.cdc = {
	
	#define CDC_DEFINITION(_idx)																	\
		[_idx] = {																					\
			.assocDescr = {																			\
				.bLength = sizeof(struct IfaceAssocDescriptor),										\
				.bDescriptorType = DESCR_TYP_ASSOC,													\
				.bFirstInterface = IF_NO_OFST_CDC_CTRL + _idx * 2,									\
				.bInterfaceCount = 2,																\
				.bFunctionClass = 2,																\
				.bFunctionSubClass = 2,																\
				.bFunctionProtocol  = 0,															\
				.iFunction = 0,																		\
			},																						\
			.ifCdcCtrl = {																			\
				.bLength = sizeof(struct UsbInterfaceDescriptor),									\
				.bDescriptorType = DESCR_TYP_IFACE,													\
				.bInterfaceNumber = IF_NO_OFST_CDC_CTRL + _idx * 2,									\
				.bNumEndpoints = 1,			/* CDC.ACM control uses 1 EP */							\
				.bInterfaceClass = 2,		/* CDC */												\
				.bInterfaceSubClass = 2,	/* ACM */												\
			},																						\
			.ifCdcHdr = {																			\
				.bFunctionLength = sizeof(struct UsbCdcHeaderFunctionalDescriptor),					\
				.bDescriptorType = DESCR_TYP_CDC_CS_IFACE,											\
				.bDescriptorSubtype = DESCR_SUBTYP_CDC_HDR,											\
				.bcdCDC = 0x0110,																	\
			},																						\
			.ifCdcAcm = {																			\
				.bFunctionLength = sizeof(struct UsbCdcAcmFunctionalDescriptor),					\
				.bDescriptorType = DESCR_TYP_CDC_CS_IFACE,											\
				.bDescriptorSubtype = DESCR_SUBTYP_CDC_ACM,											\
				.bmCapabilities = 0,																\
			},																						\
			.ifCdcUnion = {																			\
				.bFunctionLength = sizeof(struct UsbCdcUnionFunctionalDescriptor),					\
				.bDescriptorType = DESCR_TYP_CDC_CS_IFACE,											\
				.bDescriptorSubtype = DESCR_SUBTYP_CDC_UNION,										\
				.bMasterInterface = IF_NO_OFST_CDC_CTRL + _idx * 2,									\
				.bSlaveInterface0 = IF_NO_OFST_CDC_DATA + _idx * 2,									\
			},																						\
			.ifCdcCallMgmnt = {																		\
				.bFunctionLength = sizeof(struct UsbCdcCallManagementFunctionalDescriptor),			\
				.bDescriptorType = DESCR_TYP_CDC_CS_IFACE,											\
				.bDescriptorSubtype = DESCR_SUBTYP_CDC_CALL_MGMNT,									\
				.bmCapabilities = 0x03,	/* we handle shit ourselves */								\
				.bDataInterface = IF_NO_OFST_CDC_DATA + _idx * 2,									\
			},																						\
			.epCdcCtrl = {																			\
				.bLength = sizeof(struct UsbEndpointDescriptor),									\
				.bDescriptorType = DESCR_TYP_ENDPT,													\
				.bEndpointAddress = (EP_NO_OFST_CDC_CTRL + _idx * 2) | USB_DESCR_EP_NO_MASK_IN,		\
				.bmAttributes = USB_EP_DESCR_EP_TYP_INTR,											\
				.wMaxPacketSize = EP_SZ_CDC_CTRL,													\
				.bInterval = 127,																	\
			},																						\
			.ifCdcData = {																			\
				.bLength = sizeof(struct UsbInterfaceDescriptor),									\
				.bDescriptorType = DESCR_TYP_IFACE,													\
				.bInterfaceNumber = IF_NO_OFST_CDC_DATA + _idx * 2,									\
				.bNumEndpoints = 2,																	\
				.bInterfaceClass = 0x0A,	/* CDC DATA */											\
			},																						\
			.epCdcDataOut = {																		\
				.bLength = sizeof(struct UsbEndpointDescriptor),									\
				.bDescriptorType = DESCR_TYP_ENDPT,													\
				.bEndpointAddress = EP_NO_OFST_CDC_DATA_OUT + _idx * 2,								\
				.bmAttributes = USB_EP_DESCR_EP_TYP_BULK,											\
				.wMaxPacketSize = EP_SZ_CDC_DATA_OUT,												\
			},																						\
			.epCdcDataIn = {																		\
				.bLength = sizeof(struct UsbEndpointDescriptor),									\
				.bDescriptorType = DESCR_TYP_ENDPT,													\
				.bEndpointAddress = (EP_NO_OFST_CDC_DATA_IN + _idx * 2) | USB_DESCR_EP_NO_MASK_IN,	\
				.bmAttributes = USB_EP_DESCR_EP_TYP_BULK,											\
				.wMaxPacketSize = EP_SZ_CDC_DATA_IN,												\
			},																						\
		}
		
		CDC_DEFINITION(0),
	#if NUM_CDCS >= 2
		CDC_DEFINITION(1),
	#endif
	#if NUM_CDCS >= 3
		CDC_DEFINITION(2),
	#endif
		#undef CDC_DEFINITION
	},
};

static const struct UsbStringDescriptor mStrSupportedLangs = {.bLength = sizeof(struct UsbStringDescriptor) + 2 * 1, .bDescriptorType = DESCR_TYP_STRING, .chars = {0x0409},};
static const struct UsbStringDescriptor mStrManufacturer = {.bLength = sizeof(struct UsbStringDescriptor) + 2 * 9, .bDescriptorType = DESCR_TYP_STRING, .chars = {'D', 'm', 'i', 't', 'r', 'y', '.', 'G', 'R', }};
static const struct UsbStringDescriptor mStrProduct = {.bLength = sizeof(struct UsbStringDescriptor) + 2 * 9, .bDescriptorType = DESCR_TYP_STRING, .chars = {'L', 'i', 'n', 'u', 'x', 'C', 'a', 'r', 'd', }};
static const struct UsbStringDescriptor mStrSerialNumber = {.bLength = sizeof(struct UsbStringDescriptor) + 2 * 12, .bDescriptorType = DESCR_TYP_STRING, .chars = {'3', '1', '4', '1', '5', '9', '2', '6', '5', '3', '5', '8'}};

static const struct UsbStringDescriptor* mStringDescrs[] = {
	[STR_IDX_LANG_ID] = &mStrSupportedLangs,
	[STR_IDX_MANUF] = &mStrManufacturer,
	[STR_IDX_PRODNAME] = &mStrProduct,
	[STR_IDX_SNUM] = &mStrSerialNumber,
};

const void* usbExtGetGescriptor(uint_fast8_t descrType, uint_fast8_t descrIdx, uint_fast16_t langIdx, uint16_t *lenP)
{
	if (descrType == DESCR_TYP_DEVICE && descrIdx == 0) {
		
		*lenP = sizeof(mDevDescr);
		return &mDevDescr;
	}
	else if (descrType == DESCR_TYP_CONFIG && descrIdx == CFG_IDX_UARTS - 1) {
		
		*lenP = sizeof(mCfgDescr);
		return &mCfgDescr;
	}
	else if (descrType == DESCR_TYP_STRING) {
		
		if (descrIdx < sizeof(mStringDescrs) / sizeof(*mStringDescrs) && mStringDescrs[descrIdx]) {
			
			*lenP = mStringDescrs[descrIdx]->bLength;
			return mStringDescrs[descrIdx];
		}
		return NULL;
	}

	if ((descrType != 6 || descrIdx != 0))
		pr("no descr %u.%u\n", descrType, descrIdx);
	return NULL;
}

void usartSetBuadrate(uint32_t baud)
{
	
}

void __attribute__((noinline)) usartInit(void)
{
	if (!usbInit())
		pr("failed to init USB\n");
	else {

		if (!usbAttach(true))
			pr("failed to attach USB\n");
		else
			pr("usb up\n");
	}
}

void usartTxEx(uint8_t channel, uint8_t ch)
{
	static volatile uint32_t mTxByte[NUM_CDCS];	//for usb, alignment of 4 bytes is needed, so use 32-bit vals
	uint_fast8_t cdcIdx;
	uint64_t time;
	
	//allow 50ms per char on all non-primary channels, primary will wait forever, including for USB configuration
	
	//3,2,1,0 is the order
	cdcIdx = 3 - channel;
	if (cdcIdx >= NUM_CDCS)
		return;
	
	time = getTime();
	do {		
		if (mRunning && usbEpCanTx(EP_NO_OFST_CDC_DATA_IN + 2 * cdcIdx)) {
			
			mTxByte[cdcIdx] = ch;
			usbEpTx(EP_NO_OFST_CDC_DATA_IN + 2 * cdcIdx, (void*)(mTxByte + cdcIdx), 1, true);
			return;
		}
	} while (cdcIdx == 0 || getTime() - time < TICKS_PER_SECOND / 20);
}

void usartTx(uint8_t ch)
{
	return usartTxEx(3, ch);
}

bool usbExtSetIface(uint_fast16_t iface, uint_fast16_t setting)
{
	return true;
}
