/*
	(c) 2021 Dmitry Grinberg   https://dmitry.gr
	Non-commercial use only OR licensing@dmitry.gr
*/


#pragma GCC optimize ("Os")


#include <string.h>
#include "samda1e16b.h"
#include "usbDev.h"
#include "printf.h"



#define GET_STATUS				0
#define CLEAR_FEATURE			1
#define SET_FEATURE				3
#define SET_ADDRESS				5
#define GET_DESCRIPTOR			6
#define SET_DESCRIPTOR			7
#define GET_CONFIGURATION		8
#define SET_CONFIGURATION		9
#define GET_INTERFACE			10
#define SET_INTERFACE			11
#define SYNCH_FRAME				12

#define BM_REQ_IS_TO_HOST(r)	(!!((r) & 0x80))




#define BM_REQ_IS_DEVIC(r)		(((r) & 0x1F) == 0x00)
#define BM_REQ_IS_INTFC(r)		(((r) & 0x1F) == 0x01)
#define BM_REQ_IS_ENDPT(r)		(((r) & 0x1F) == 0x02)
#define BM_REQ_IS_OTHER(r)		(((r) & 0x1F) == 0x03)

#define BM_REQ_IS_STD(r)		(((r) & 0x60) == 0x00)
#define BM_REQ_IS_CLS(r)		(((r) & 0x60) == 0x20)
#define BM_REQ_IS_VEN(r)		(((r) & 0x60) == 0x40)

#define EP_SZ_TO_EPSZ_VAL(x)	(28 - __builtin_clz(x))

static enum UsbState mUsbState;
static uint8_t mCfgIdx, mEpSz[USB_MAX_EP_IDX + 1][2], mAlreadyNotifiedRx;
static volatile uint8_t mEp0bufRx[EP0_SZ];
static uint16_t mEp0txLeft;
static const uint8_t *mEp0TxBuf;

static volatile UsbDeviceDescBank mEpCfgs[USB_MAX_EP_IDX + 1][2] = {};	//always 2 banks per EP :(



static void usbPrvSetState(enum UsbState newState)
{
	mUsbState = newState;
	usbExtStateChangedNotif(newState);
}

static bool usbPrvEpStallNocheck(uint_fast8_t epNo, uint_fast8_t structIdx, bool stall)
{
	if (stall)
		USB->DEVICE.DeviceEndpoint[epNo].EPSTATUSSET.reg = structIdx ? USB_DEVICE_EPSTATUSSET_STALLRQ1 : USB_DEVICE_EPSTATUSSET_STALLRQ0;
	else
		USB->DEVICE.DeviceEndpoint[epNo].EPSTATUSCLR.reg = structIdx ? USB_DEVICE_EPSTATUSCLR_STALLRQ1 : USB_DEVICE_EPSTATUSCLR_STALLRQ0;
	
	return true;
}

static bool usbPrvEpStall(uint_fast8_t epNo, uint_fast8_t structIdx, bool stall)
{
	if (!epNo || epNo > USB_MAX_EP_IDX)
		return false;
	
	return usbPrvEpStallNocheck(epNo, structIdx, stall);
}

static void usbPrvControlStall(bool stall)
{
	usbPrvEpStallNocheck(0, 1, stall);
}

static bool usbPrvEpTx(uint_fast8_t epNo, const void *data, uint_fast16_t len, bool autoZlp)
{
	if (USB->DEVICE.DeviceEndpoint[epNo].EPSTATUS.bit.BK1RDY) {
		pr("sending on ep%u while busy\n", epNo);
		return false;
	}
	
	mEpCfgs[epNo][1].ADDR.reg = (uintptr_t)data;
	mEpCfgs[epNo][1].PCKSIZE.reg = USB_DEVICE_PCKSIZE_SIZE(mEpSz[epNo][1]) |
									USB_DEVICE_PCKSIZE_BYTE_COUNT(len) | 
									(autoZlp ? USB_DEVICE_PCKSIZE_AUTO_ZLP : 0);
	
	asm volatile("":::"memory");
	USB->DEVICE.DeviceEndpoint[epNo].EPINTFLAG.reg = USB_DEVICE_EPINTFLAG_TRCPT1 | USB_DEVICE_EPINTFLAG_TRFAIL1 | USB_DEVICE_EPINTFLAG_STALL1;
	USB->DEVICE.DeviceEndpoint[epNo].EPSTATUSSET.reg = USB_DEVICE_EPSTATUSSET_BK1RDY;

	return true;
}

static void usbPrvSendNextEp0(void)
{
	static uint8_t __attribute__((aligned(4))) buf[EP0_SZ];
	uint_fast16_t now = mEp0txLeft;
	
	if (now > EP0_SZ)
		now = EP0_SZ;
	
	memcpy(buf, mEp0TxBuf, now);
	if (mEp0txLeft -= now)
		mEp0TxBuf += now;
	else
		mEp0TxBuf = NULL;
	
	usbPrvEpTx(0, buf, now, false);
}

static void usbPrvScheduleTxEp0(const void *data, uint_fast16_t len)
{
	if (mEp0TxBuf) {
		
		pr("cannot send on EP0 while already sending\n");
		return;
	}
	
	asm volatile("":::"memory");
	mEp0TxBuf = data;
	mEp0txLeft = len;
	usbPrvSendNextEp0();
}


bool usbEpTx(uint_fast8_t epNo, const void *data, uint_fast16_t len, bool autoZlp)
{
	if (!epNo || epNo > USB_MAX_EP_IDX)
		return false;
	
	if (len >= 16384)
		return false;
	
	return usbPrvEpTx(epNo, data, len, autoZlp);
}

bool usbEpCanTx(uint_fast8_t epNo)
{
	if (!epNo || epNo > USB_MAX_EP_IDX)
		return false;
	
	return !USB->DEVICE.DeviceEndpoint[epNo].EPSTATUS.bit.BK1RDY;
}

static void usbPrvControlSend(const void *data, uint_fast16_t len)
{
	usbPrvControlStall(false);
	usbPrvScheduleTxEp0(data, len);
}

static bool usbPrvHandleGetDescriptor(uint_fast8_t descrType, uint_fast8_t descrIdx, uint_fast16_t langIdx, uint_fast16_t maxLen)
{
	const void *descr;
	uint16_t descrLen;
	
	descr = usbExtGetGescriptor(descrType, descrIdx, langIdx, &descrLen);
	if (descr) {
		if (descrLen > maxLen)
			descrLen = maxLen;
		
		usbPrvControlSend(descr, descrLen);
		return true;
	}
	return false;
}

bool usbPrvEpIsStalledNoCheck(uint_fast8_t epNo, uint_fast8_t structIdx)
{
	uint32_t sta = USB->DEVICE.DeviceEndpoint[epNo].EPSTATUS.reg;
	
	return !!(sta & (structIdx ? USB_DEVICE_EPSTATUS_STALLRQ1 : USB_DEVICE_EPSTATUS_STALLRQ0));
}

uint_fast8_t usbEp0dataRx(uint8_t *data)
{
	uint_fast8_t len;
	
	while (!USB->DEVICE.DeviceEndpoint[0].EPSTATUS.bit.BK0RDY);
	asm volatile("":::"memory");
	len = mEpCfgs[0][0].PCKSIZE.bit.BYTE_COUNT;
	memcpy(data, (void*)mEp0bufRx, len);
	asm volatile("":::"memory");
	mEpCfgs[0][0].PCKSIZE.reg = USB_DEVICE_PCKSIZE_SIZE(mEpSz[0][0]);
	USB->DEVICE.DeviceEndpoint[0].EPSTATUSCLR.reg = USB_DEVICE_EPSTATUSSET_BK0RDY;
	return len;
}

static void usbPrvHandleSetup(void)
{
	struct UsbSetup *pSetup = (struct UsbSetup*)mEp0bufRx, setup = *pSetup;
	uint_fast8_t rxLen = mEpCfgs[0][0].PCKSIZE.bit.BYTE_COUNT;
	static uint32_t mTxBuf[2];	//a small buffer for small TXs
	void *cliDataP = mTxBuf;
	int16_t cliSz;
	
	asm volatile("":::"memory");
	mEpCfgs[0][0].PCKSIZE.reg = USB_DEVICE_PCKSIZE_SIZE(mEpSz[0][0]);
	USB->DEVICE.DeviceEndpoint[0].EPSTATUSCLR.reg = USB_DEVICE_EPSTATUSSET_BK0RDY;
	
	if (BM_REQ_IS_STD(setup.bmRequestType)) {
		
		if (BM_REQ_IS_TO_HOST(setup.bmRequestType) && setup.bRequest == GET_DESCRIPTOR) {
			
			//wValueH is descriptor type
			//wValueL is descriptor index (only for config & string descriptors, else zero)
			//wIndex is languageID for string, zero else
			//wLength is max num bytes to return
			if (usbPrvHandleGetDescriptor(setup.wValueH, setup.wValueL, setup.wIndex, setup.wLength))
				goto success;
			
			goto stall_quietly;
		}
		if (!BM_REQ_IS_TO_HOST(setup.bmRequestType) && setup.bRequest == SET_CONFIGURATION){
		
			//wValueL is desired config
			mCfgIdx = setup.wValueL;
			usbExtConfigSelected(setup.wValueL);
			usbPrvSetState(UsbRunning);
			usbPrvControlSend(&setup, 0);	//send ZLP
			goto success;
		}
		if (!BM_REQ_IS_TO_HOST(setup.bmRequestType) && setup.bRequest == SET_ADDRESS){
		
			//wValueL is address
			usbPrvSetState(UsbAssigningAddress);
			USB->DEVICE.DADD.reg = USB_DEVICE_DADD_DADD(setup.wValueL);
			usbPrvControlSend(&setup, 0);	//send ZLP
			goto success;
		}
		if (BM_REQ_IS_TO_HOST(setup.bmRequestType) && setup.bRequest == GET_CONFIGURATION){
		
			mTxBuf[0] = mCfgIdx;
			usbPrvControlSend(mTxBuf, 1);
			goto success;
		}
		if (!BM_REQ_IS_TO_HOST(setup.bmRequestType) && BM_REQ_IS_INTFC(setup.bmRequestType) && setup.bRequest == SET_INTERFACE) {
			
			if (usbExtSetIface(setup.wIndex, setup.wValue)) {
				usbPrvControlSend(&setup, 0);	//send ZLP
				goto success;
			}
			else
				goto stall_quietly;
		}
		if (BM_REQ_IS_ENDPT(setup.bmRequestType) && (setup.bRequest == CLEAR_FEATURE || setup.bRequest == SET_FEATURE)) {	//clear stall
				
			bool in = !!(setup.wIndexL & 0x80);
			uint_fast8_t epNo = setup.wIndexL & 0x7f;
			
			if (in)
				usbEpStallIn(epNo, setup.bRequest == SET_FEATURE);
			else
				usbEpStallOut(epNo, setup.bRequest == SET_FEATURE);
			
			goto success;
		}
		if (BM_REQ_IS_TO_HOST(setup.bmRequestType) && setup.bRequest == GET_STATUS && setup.wLength == 2) {
				
			if (BM_REQ_IS_ENDPT(setup.bmRequestType)) {
				
				mTxBuf[0] = usbPrvEpIsStalledNoCheck(setup.wIndexL & 0x7f, (setup.wIndexL & 0x80) ? 1 : 0) ? 1 : 0;
				usbPrvControlSend(mTxBuf, 2);
				goto success;
			}
			else if (BM_REQ_IS_INTFC(setup.bmRequestType) || BM_REQ_IS_DEVIC(setup.bmRequestType)) {
				
				//different meanings, but for us all zeros
				mTxBuf[0] = 0;
				usbPrvControlSend(mTxBuf, 2);
				goto success;
			}
		}
	}
	else if (usbExtEp0handler(&setup, &cliDataP, &cliSz)) {
		
		if (cliSz < 0)
			goto stall_quietly;
		else {
			usbPrvControlSend(cliDataP, cliSz);
			goto success;
		}
	}

	//printing this causes issues since it is a large delay. comment out when debugging
	if (1) {
		pr("SETUP %u bytes\n", rxLen);
		pr("unknown request (%u b) %02x.%02x, 0x%04x 0x%04x 0x%04x\n", rxLen, setup.bmRequestType, setup.bRequest,
			setup.wValue, setup.wIndex, setup.wLength);
		pr("HALTED\n");
		while(1);
	}
	
stall_quietly:
	usbPrvControlStall(true);
	return;

success:
	return;
}

enum UsbState usbGetState(void)
{
	return mUsbState;
}

void __attribute__((used)) USB_IRQHandler(void)
{
	uint32_t sta = USB->DEVICE.INTFLAG.reg, clr = 0;
	uint32_t ep0int = USB->DEVICE.DeviceEndpoint[0].EPINTFLAG.reg;
	uint_fast8_t i;
	
	if (sta & USB_DEVICE_INTFLAG_EORST) {
	
		mEp0TxBuf = NULL;
		USB->DEVICE.DADD.reg = 0;
		clr += USB_DEVICE_INTENCLR_EORST;
		USB->DEVICE.DeviceEndpoint[0].EPCFG.reg = USB_DEVICE_EPCFG_EPTYPE0(1) | USB_DEVICE_EPCFG_EPTYPE1(1);
		USB->DEVICE.DeviceEndpoint[0].EPINTENSET.reg = USB_DEVICE_EPINTENSET_TRCPT0 | USB_DEVICE_EPINTENSET_RXSTP | USB_DEVICE_EPINTENSET_TRCPT1;
		mCfgIdx = 0;
		usbPrvSetState(UsbWaitForEnum);
	}
	else {
	
		if (ep0int & USB_DEVICE_EPINTFLAG_RXSTP) {
			
			USB->DEVICE.DeviceEndpoint[0].EPINTFLAG.reg = USB_DEVICE_EPINTFLAG_RXSTP;
			ep0int &=~ USB_DEVICE_EPINTFLAG_RXSTP;
			usbPrvHandleSetup();
		}
		if (ep0int & USB_DEVICE_EPINTFLAG_TRCPT0) {
			
			USB->DEVICE.DeviceEndpoint[0].EPINTFLAG.reg = USB_DEVICE_EPINTFLAG_TRCPT0;
			ep0int &=~ USB_DEVICE_EPINTFLAG_TRCPT0;
		}
		if (ep0int & USB_DEVICE_EPINTFLAG_TRCPT1) {
			
			//clear it right away so we dont miss the next one
			USB->DEVICE.DeviceEndpoint[0].EPINTFLAG.reg = USB_DEVICE_EPINTFLAG_TRCPT1;
			
			if (mUsbState == UsbAssigningAddress) {
				
				USB->DEVICE.DADD.bit.ADDEN = 1;
				usbPrvSetState(UsbWaitingForConfig);
			}
			
			ep0int &=~ USB_DEVICE_EPINTFLAG_TRCPT1;
			usbPrvSendNextEp0();
		}
		
		if (ep0int &~ (ep0int | USB_DEVICE_EPINTFLAG_STALL1)) {
			
			pr("usb unhandled irq EPINTFLAG = 0x%04x, ep0sta = 0x%08x\n", ep0int);
			USB->DEVICE.DeviceEndpoint[0].EPINTFLAG.reg = ep0int;
		}
		
		for (i = 1; i <= USB_MAX_EP_IDX; i++) {
			
			if (USB->DEVICE.DeviceEndpoint[i].EPSTATUS.bit.BK0RDY && !(mAlreadyNotifiedRx & (1 << i))) {
				
				mAlreadyNotifiedRx |= 1 << i;
				usbExtEpDataArrivalNotif(i, mEpCfgs[i][0].PCKSIZE.bit.BYTE_COUNT);
			}
			USB->DEVICE.DeviceEndpoint[i].EPINTFLAG.reg = USB_DEVICE_EPINTFLAG_TRCPT0;
		}
	}
	USB->DEVICE.INTFLAG.reg = clr;
}

bool usbRxRelease(uint_fast8_t epNo)
{
	if (!epNo || epNo > USB_MAX_EP_IDX)
		return false;
	
	if (!USB->DEVICE.DeviceEndpoint[epNo].EPSTATUS.bit.BK0RDY)
		pr("releaseing unused rx buffer\n");
	
	mAlreadyNotifiedRx &=~ (1 << epNo);
	asm volatile("":::"memory");
	mEpCfgs[epNo][0].PCKSIZE.reg = USB_DEVICE_PCKSIZE_SIZE(mEpSz[epNo][0]);
	USB->DEVICE.DeviceEndpoint[epNo].EPSTATUSCLR.reg = USB_DEVICE_EPSTATUSSET_BK0RDY;
	
	return true;
}

static bool usbPrvEpUnconfig(uint_fast8_t epNo, uint_fast8_t structIdx)
{
	if (!epNo || epNo > USB_MAX_EP_IDX)
		return false;
	
	if (structIdx)		//IN
		USB->DEVICE.DeviceEndpoint[epNo].EPCFG.bit.EPTYPE1 = 0;
	else				//OUT
		USB->DEVICE.DeviceEndpoint[epNo].EPCFG.bit.EPTYPE0 = 0;
	
	return true;
}

static bool usbPrvEpCfg(uint_fast8_t epNo, uint_fast16_t epSz, void *dataPtr, uint_fast8_t epType, uint_fast8_t structIdx, bool makeAvailable)
{
	if (!epSz || epSz > 64 || (epSz & (epSz - 1)))
		return false;
	
	if (!epNo || epNo > USB_MAX_EP_IDX)
		return false;
	
	pr("ep %u.%u sz %u type %u\n", epNo, structIdx, epSz, epType);
	epSz = EP_SZ_TO_EPSZ_VAL(epSz);
	
	mEpSz[epNo][structIdx] = epSz;
	
	mEpCfgs[epNo][structIdx].ADDR.reg = (uintptr_t)dataPtr;
	mEpCfgs[epNo][structIdx].PCKSIZE.reg = USB_DEVICE_PCKSIZE_SIZE(epSz);
	
	if (structIdx) {	//IN
		USB->DEVICE.DeviceEndpoint[epNo].EPCFG.bit.EPTYPE1 = epType;
	}
	else {				//OUT
		USB->DEVICE.DeviceEndpoint[epNo].EPCFG.bit.EPTYPE0 = epType;
		USB->DEVICE.DeviceEndpoint[epNo].EPINTENSET.reg = USB_DEVICE_EPINTENSET_TRCPT0;
	}
	
	return true;
}

static bool usbPrvEpIsStalled(uint_fast8_t epNo, uint_fast8_t structIdx)
{
	if (!epNo || epNo > USB_MAX_EP_IDX)
		return false;
	
	return usbPrvEpIsStalledNoCheck(epNo, structIdx);
}

bool usbEpIsStalledIn(uint_fast8_t epNo)
{
	return usbPrvEpIsStalled(epNo, 1);
}

bool usbEpIsStalledOut(uint_fast8_t epNo)
{
	return usbPrvEpIsStalled(epNo, 0);
}

bool usbEpStallIn(uint_fast8_t epNo, bool stall)
{
	return usbPrvEpStall(epNo, 1, stall);
}

bool usbEpStallOut(uint_fast8_t epNo, bool stall)
{
	return usbPrvEpStall(epNo, 0, stall);
}

bool usbEpUnconfigIn(uint_fast8_t epNo)
{
	return usbPrvEpUnconfig(epNo, 1);
}

bool usbEpUnconfigOut(uint_fast8_t epNo)
{
	return usbPrvEpUnconfig(epNo, 0);
}

bool usbEpIntrCfgIn(uint_fast8_t epNo, uint_fast16_t epSz)
{
	return usbPrvEpCfg(epNo, epSz, NULL, 4, 1, false);
}

bool usbEpIntrCfgOut(uint_fast8_t epNo, uint_fast16_t epSz, void *dataPtr)
{
	return usbPrvEpCfg(epNo, epSz, dataPtr, 4, 0, true);
}

bool usbEpBulkCfgIn(uint_fast8_t epNo, uint_fast16_t epSz)
{
	return usbPrvEpCfg(epNo, epSz, NULL, 3, 1, false);
}

bool usbEpBulkCfgOut(uint_fast8_t epNo, uint_fast16_t epSz, void *dataPtr)
{
	return usbPrvEpCfg(epNo, epSz, dataPtr, 3, 0, true);
}

bool usbInit(void)
{
	usbPrvSetState(UsbDisconnected);
	
	if (EP0_SZ < 8 || EP0_SZ > 64 || (EP0_SZ & (EP0_SZ - 1)))
		return false;
	
	mEpCfgs[0][0].ADDR.reg = (uintptr_t)mEp0bufRx,
	mEpCfgs[0][0].PCKSIZE.reg = USB_DEVICE_PCKSIZE_SIZE(EP_SZ_TO_EPSZ_VAL(EP0_SZ));
	
	mEpSz[0][0] = EP_SZ_TO_EPSZ_VAL(EP0_SZ);
	mEpSz[0][1] = EP_SZ_TO_EPSZ_VAL(EP0_SZ);
	
	//reset
	USB->DEVICE.CTRLA.bit.SWRST = 1;
	while (USB->DEVICE.SYNCBUSY.bit.SWRST);
	USB->DEVICE.CTRLA.bit.SWRST = 0;
	while (USB->DEVICE.SYNCBUSY.bit.SWRST);
	
	//configs
	USB->DEVICE.CTRLA.reg = USB_CTRLA_RUNSTDBY | USB_CTRLA_MODE_DEVICE;
	USB->DEVICE.CTRLB.reg = USB_DEVICE_CTRLB_DETACH | USB_DEVICE_CTRLB_SPDCONF_FS | USB_DEVICE_CTRLB_LPMHDSK_NO;
	USB->DEVICE.DESCADD.reg = (uintptr_t)mEpCfgs;
	USB->DEVICE.DADD.reg = 0;
	USB->DEVICE.INTENSET.reg = USB_DEVICE_INTENSET_EORST;
	USB->DEVICE.CTRLA.bit.ENABLE = 1;
	
	NVIC_EnableIRQ(USB_IRQn);
	
	return true;
}

bool usbAttach(bool attached)
{
	if (attached) {
		if (mUsbState != UsbDisconnected)
			return false;
		
		usbPrvSetState(UsbWaitForReset);
		USB->DEVICE.CTRLB.bit.DETACH = 0;
		return true;
	}
	else if (mUsbState == UsbDisconnected)
		return false;
	else {
		
		USB->DEVICE.CTRLB.bit.DETACH = 1;
		usbPrvSetState(UsbDisconnected);
		return true;
	}
}

