/*
	(c) 2021 Dmitry Grinberg   https://dmitry.gr
	Non-commercial use only OR licensing@dmitry.gr
*/

#include <stdbool.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include "partitions.h"
#include "printf.h"
#include "entry.h"
#include "util.h"
#include "load.h"


#define MY_SYS_TYPE			0x82012001	//cpu=0x82 (r3000), machine is DS2100/3100
#define STRINGIFY2(x)		#x
#define STRINGIFY(x)		STRINGIFY2(x)
#define MY_SYS_TYPE_STR		STRINGIFY(MY_SYS_TYPE)


void __attribute__((used)) prom_restart_reboot_halt(void)
{
	pr("\r\n\r\nHALTED\r\n\r\n");
	while(1);
}

int __attribute__((used)) prom_getchar(void)
{
	return -1;
}

void __attribute__((used)) prom_putchar(char chr)
{
	consoleWrite(chr);
}

const char* __attribute__((used)) prom_getenv(const char *envVar)
{
	static const char *env[] = {
		"systype", MY_SYS_TYPE_STR,
		"osconsole", "0,0",				//ultrix: grapical console is anything with the lowest bit set in first byte. netbsd needs "1,0" here for gfx boot, "3" for serial on last serial port
		"scsiid0", "7",
	#ifdef REPORT_BOOT_FROM_CD
		"boot", "rz(0,0,0)",
	#else
		"boot", "rz(0,6,0)",
	#endif
		"haltaction", "b",				//b for reboot h for halt
		NULL
	};
	const char **envp = env;
	
	for (envp = env; envp[0]; envp += 2) {
		if (!strcmp(envVar, envp[0]))
			return envp[1];
	}
	
	fatal("got asked for unknown env type '%s', not sure how to reply", envVar);
}

void prPutchar(char chr)
{
	consoleWrite(chr);
}

static void* v2p(void* addr)
{
	return (void*)(((uintptr_t)addr) & 0x1fffffff);	//assumes a lot of things :D
}

static bool diskReadForFs(void *userData, uint32_t sec, void *dst)
{
	uint32_t startSec = (uint32_t)userData;
	
	return readblock(sec + startSec, v2p(dst));
}

static bool diskReadForPartMgr(void *userData, void *dst, uint32_t sector)
{
	(void)userData;
	
	return readblock(sector, v2p(dst));
}

void __attribute__((noreturn)) start(void)
{
	static uint8_t __attribute__((aligned(512))) mMiscDiscBuf[VFS_SECTOR_SIZE];	//align guarantees no kilobyte crossings for sector read
	struct VfsFileOpenInfoOpaque foi;
	char fName[VFS_MAX_NAME_LEN + 1];
	uint32_t startSec, numSec;
	struct VfsFile *dir, *fil;
	struct VfsVolume *vol;
	
	pr("hello, world\n");

	if (!partMgrFindBootPartition(diskReadForPartMgr, NULL, mMiscDiscBuf, &startSec, &numSec))
		fatal("no boot partition\n");
	
	pr("Boot partition found at %u + %u sec\n", startSec, numSec);
	
	vfsInit(mMiscDiscBuf);
	
	vol = vfsMount(diskReadForFs, (void*)startSec);
	if (!vol)
		fatal("mount failed\n");
	
	dir = vfsOpenRoot(vol);
	if (!dir)
		fatal("failed to open root dir\n");

	while (vfsDirEnum(dir, fName, &foi)) {
		
		uint32_t fileSize;
		uint8_t type;
		
		pr(" > '%s'\n", fName);
		
		if (!isValidKernelName(fName) || !vfsFileInfo(vol, &foi, &type, &fileSize) || type != VFS_TYPE_FILE || fileSize < (1 << 20))
			continue;
		
		fil = vfsFileOpen(vol, &foi);
		if (!fil)
			pr("cannot open file\n");
		else {
			
			loadOS(fil, fileSize, 0, NULL);
			pr("failed to load this kernel\n");
			vfsFileClose(fil);
		}
	}
	vfsFileClose(dir);
	
	fatal("No loadable kernels found\n");
}
